@@ -15,6 +15,7 @@ from typing import Mapping, MutableMapping, Optional, Sequence, Tuple, Union
 
 import jinja2
 import tomlkit
+import tomlkit.items
 from dunamai import (
     bump_version,
     check_version,
@@ -86,6 +87,14 @@ if sys.version_info >= (3, 8):
         },
     )
 
+    _Bump = TypedDict(
+        "_Bump",
+        {
+            "enable": bool,
+            "index": int,
+        },
+    )
+
     _Config = TypedDict(
         "_Config",
         {
@@ -102,7 +111,7 @@ if sys.version_info >= (3, 8):
             "format": Optional[str],
             "format-jinja": Optional[str],
             "format-jinja-imports": Sequence[_JinjaImport],
-            "bump": bool,
+            "bump": Union[bool, _Bump],
             "tagged-metadata": bool,
             "full-commit": bool,
             "tag-branch": Optional[str],
@@ -110,13 +119,21 @@ if sys.version_info >= (3, 8):
             "strict": bool,
             "fix-shallow-repository": bool,
             "ignore-untracked": bool,
+            "commit-length": Optional[int],
+            "commit-prefix": Optional[str],
+            "escape-with": Optional[str],
             "from-file": _FromFile,
         },
     )
 else:
-
-    class _Config(Mapping):
-        pass
+    _SubstitutionPattern = Mapping
+    _SubstitutionFolder = Mapping
+    _Substitution = Mapping
+    _File = Mapping
+    _JinjaImport = Mapping
+    _FromFile = Mapping
+    _Bump = Mapping
+    _Config = Mapping
 
 
 class _Mode(Enum):
@@ -131,14 +148,14 @@ class _ProjectState:
         original_version: Optional[str],
         version: str,
         mode: _Mode,
-        dynamic_index: Optional[int],
+        dynamic_array: Optional[tomlkit.items.Array],
         substitutions: Optional[MutableMapping[Path, str]] = None,
     ) -> None:
         self.path = path
         self.original_version = original_version
         self.version = version
         self.mode = mode
-        self.dynamic_index = dynamic_index
+        self.dynamic_array = dynamic_array
         self.substitutions = {} if substitutions is None else substitutions  # type: MutableMapping[Path, str]
 
 
@@ -194,6 +211,19 @@ class _FolderConfig:
         return [main, *extra]
 
 
+class _BumpConfig:
+    def __init__(self, enable: bool, index: int):
+        self.enable = enable
+        self.index = index
+
+    @staticmethod
+    def from_config(config: Union[bool, _Bump]) -> "_BumpConfig":
+        if isinstance(config, bool):
+            return _BumpConfig(config, -1)
+        else:
+            return _BumpConfig(config["enable"], config["index"])
+
+
 def _default_config() -> Mapping:
     return {
         "tool": {
@@ -229,6 +259,9 @@ def _default_config() -> Mapping:
                 "strict": False,
                 "fix-shallow-repository": False,
                 "ignore-untracked": False,
+                "commit-length": None,
+                "commit-prefix": None,
+                "escape-with": None,
                 "from-file": {
                     "source": None,
                     "pattern": None,
@@ -368,8 +401,18 @@ def _render_jinja(version: Version, template: str, config: _Config, extra: Optio
     if extra is None:
         extra = {}
 
-    if config["bump"] and version.distance > 0:
-        version = version.bump()
+    bump_config = _BumpConfig.from_config(config["bump"])
+
+    if bump_config.enable and version.distance > 0:
+        version = version.bump(index=bump_config.index)
+
+    def base_part(index: int) -> int:
+        parts = version.base.split(".")
+        try:
+            return int(parts[index])
+        except Exception:
+            return 0
+
     default_context = {
         "base": version.base,
         "version": version,
@@ -381,6 +424,9 @@ def _render_jinja(version: Version, template: str, config: _Config, extra: Optio
         "branch": version.branch,
         "branch_escaped": _escape_branch(version.branch),
         "timestamp": _format_timestamp(version.timestamp),
+        "major": base_part(0),
+        "minor": base_part(1),
+        "patch": base_part(2),
         "env": os.environ,
         "bump_version": bump_version,
         "tagged_metadata": version.tagged_metadata,
@@ -469,6 +515,7 @@ def _get_version_from_dunamai(
         strict=config["strict"] if strict is None else strict,
         pattern_prefix=config["pattern-prefix"],
         ignore_untracked=config["ignore-untracked"],
+        commit_length=config["commit-length"],
     )
 
 
@@ -508,13 +555,20 @@ def _get_version(config: _Config, name: Optional[str] = None) -> Tuple[str, Vers
         if style is not None:
             check_version(serialized, style)
     else:
-        serialized = version.serialize(
+        bump_config = _BumpConfig.from_config(config["bump"])
+        if bump_config.enable:
+            updated = version.bump(index=bump_config.index, smart=True)
+        else:
+            updated = version
+
+        serialized = updated.serialize(
             metadata=config["metadata"],
             dirty=config["dirty"],
             format=config["format"],
             style=style,
-            bump=config["bump"],
             tagged_metadata=config["tagged-metadata"],
+            commit_prefix=config["commit-prefix"],
+            escape_with=config["escape-with"],
         )
 
     return (serialized, version)
@@ -593,8 +647,11 @@ def _apply_version(
     if mode == _Mode.Classic:
         pyproject["tool"]["poetry"]["version"] = version  # type: ignore
     elif mode == _Mode.Pep621:
-        pyproject["project"]["dynamic"].remove("version")  # type: ignore
+        if "version" in pyproject["project"]["dynamic"]:  # type: ignore
+            pyproject["project"]["dynamic"].remove("version")  # type: ignore
         pyproject["project"]["version"] = version  # type: ignore
+        if "version" in pyproject["tool"]["poetry"]:  # type: ignore
+            pyproject["tool"]["poetry"].pop("version")  # type: ignore
 
     # Disable the plugin in case we're building a source distribution,
     # which won't have access to the VCS info at install time.
@@ -633,7 +690,6 @@ def _apply_version(
 
 
 def _get_and_apply_version(
-    pyproject: Optional[Mapping] = None,
     pyproject_path: Optional[Path] = None,
     retain: bool = False,
     force: bool = False,
@@ -644,8 +700,9 @@ def _get_and_apply_version(
         if pyproject_path is None:
             raise RuntimeError("Unable to find pyproject.toml")
 
-    if pyproject is None:
-        pyproject = tomlkit.parse(pyproject_path.read_bytes().decode("utf-8"))
+    # The actual type is `tomlkit.TOMLDocument`, which is important to preserve formatting,
+    # but it also causes a lot of type-checking noise.
+    pyproject = tomlkit.parse(pyproject_path.read_bytes().decode("utf-8"))  # type: Mapping
 
     classic = "tool" in pyproject and "poetry" in pyproject["tool"] and "name" in pyproject["tool"]["poetry"]
     pep621 = (
@@ -654,18 +711,21 @@ def _get_and_apply_version(
         and "dynamic" in pyproject["project"]
         and "version" in pyproject["project"]["dynamic"]
         and "version" not in pyproject["project"]
+        and "tool" in pyproject
+        and "poetry" in pyproject["tool"]
+        and "version" in pyproject["tool"]["poetry"]
     )
 
     if classic:
         name = pyproject["tool"]["poetry"]["name"]
         original = pyproject["tool"]["poetry"]["version"]
-        dynamic_index = None
+        dynamic_array = None
     elif pep621:
         name = pyproject["project"]["name"]
-        original = pyproject["project"].get("version")
-        dynamic_index = pyproject["project"]["dynamic"].index("version")
+        original = pyproject["tool"]["poetry"]["version"]
+        dynamic_array = pyproject["project"]["dynamic"]
     else:
-        return name if name in _state.projects else None
+        return None
 
     if name in _state.projects:
         return name
@@ -684,12 +744,12 @@ def _get_and_apply_version(
 
     if classic and name is not None and original is not None:
         mode = _Mode.Classic
-        _state.projects[name] = _ProjectState(pyproject_path, original, version, mode, dynamic_index)
+        _state.projects[name] = _ProjectState(pyproject_path, original, version, mode, dynamic_array)
         if io:
             _apply_version(name, version, instance, config, pyproject_path, mode, retain)
     elif pep621 and name is not None:
         mode = _Mode.Pep621
-        _state.projects[name] = _ProjectState(pyproject_path, original, version, mode, dynamic_index)
+        _state.projects[name] = _ProjectState(pyproject_path, original, version, mode, dynamic_array)
         if io:
             _apply_version(name, version, instance, config, pyproject_path, mode, retain)
 
@@ -706,7 +766,7 @@ def _revert_version(retain: bool = False) -> None:
             persistent = []
             for file, file_info in config["files"].items():
                 if file_info["persistent-substitution"]:
-                    persistent.append(state.path.parent.joinpath(file))
+                    persistent.append(state.path.parent.joinpath(file).resolve())
 
             for file, content in state.substitutions.items():
                 if file in persistent:
@@ -721,11 +781,12 @@ def _revert_version(retain: bool = False) -> None:
             if state.original_version is not None:
                 pyproject["tool"]["poetry"]["version"] = state.original_version  # type: ignore
         elif state.mode == _Mode.Pep621:
-            if state.dynamic_index is not None:
-                index = state.dynamic_index
-                pyproject["project"]["dynamic"].insert(index, "version")  # type: ignore
+            if state.dynamic_array is not None:
+                pyproject["project"]["dynamic"] = state.dynamic_array  # type: ignore
             if "version" in pyproject["project"]:  # type: ignore
                 pyproject["project"].pop("version")  # type: ignore
+            if state.original_version is not None:
+                pyproject["tool"]["poetry"]["version"] = state.original_version  # type: ignore
 
         if not retain and not _state.cli_mode:
             pyproject["tool"]["poetry-dynamic-versioning"]["enable"] = True  # type: ignore
